import enum


class Command(enum.Enum):
    """Enumeration of RBF commands."""

    TOGGLE = "*"
    """Toggle the current bit."""
    TAPE_RIGHT = ">"
    """Shift the tape head right."""

    TAPE_LEFT = "<"
    """Shift the tape head left."""

    LOOP_START = "("
    """If the current bit is zero, jump past matching ). Else, continue (loop)."""

    LOOP_END = ")"
    """If the current bit is zero, jump back to just after matching (. Else, continue."""



from rbf_lang.runner import run


def test_move_bit() -> None:
    move_right = """
    # Move the bit under the pointer to the right, using a temporary cell
    # x=?, y=0, f=0
    (>>*<<) # set f if x is set
    >>( # if f is set
        <(>*<)* # set y
        <*(>>*<<) # unset x
    >>)
    <(>*<) # if y is set, unset f
    (>>*<<)>>(<(>*<)*<*(>>*<<)>>)<(>*<)
    (>>*<<)>>(<(>*<)*<*(>>*<<)>>)<(>*<)
    (>>*<<)>>(<(>*<)*<*(>>*<<)>>)<(>*<)
    """

    program, tape = run(move_right, "100")

    assert program.steps == 100
    assert tape == "010"
    program.reset()
    program, tape = run(program, tape)

    assert program.steps == 100
    assert tape == "001"



import pytest
from typing import Any, Generator
from pathlib import Path
import sys


__project_root__ = Path(__file__).resolve().parents[1]
sys.path.append(str(__project_root__))


@pytest.fixture(autouse=True)
def add_space_before_first_print() -> Generator[None, None, None]:
    import builtins

    _print = builtins.print

    def patched_print(*args: Any, **kwargs: Any) -> None:
        _print(end="\n")  # Add a newline
        _print(*args, **kwargs)
        builtins.print = _print

    builtins.print = patched_print
    try:
        yield
    finally:
        builtins.print = _print



import warnings

from typing import Sequence, Union, overload, Optional
from .command import Command
from .exceptions import InvalidProgramError, ProgramPointerError

_ProgramInitType = Union[str, Sequence[Command], "Program"]


class Program(Sequence[Command]):
    """RBF program."""

    _program: Sequence[Command]
    _pointer: int
    _steps: int

    def __init__(
        self,
        program: _ProgramInitType,
        pointer: Optional[int] = None,
    ) -> None:
        if isinstance(program, Program):
            self._program = program.program
            self._pointer = program.pointer
            self._steps = program.steps
            if pointer is not None:
                warnings.warn(
                    "Pointer argument is ignored when initializing Program with another Program. Set it to None to disable this warning.",
                    stacklevel=2,
                )
        elif isinstance(program, str) or isinstance(program, Sequence):
            if isinstance(program, str):
                program = preprocess_program(program)
            validated_program = validate_program(program)
            self._program = validated_program
            pointer = 0 if pointer is None else pointer
            self._pointer = pointer
            self._steps = 0
        else:
            raise TypeError("Program must be initialized with a string or a sequence.")

    @property
    def program(self) -> Sequence[Command]:
        return self._program

    @property
    def pointer(self) -> int:
        return self._pointer

    @property
    def steps(self) -> int:
        return self._steps

    def __len__(self) -> int:
        return len(self._program)

    @overload
    def __getitem__(self, index: int) -> Command: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[Command]: ...

    def __getitem__(
        self,
        index_or_slice: Union[int, slice],
    ) -> Union[Command, Sequence[Command]]:
        if isinstance(index_or_slice, int):
            return self._program[index_or_slice]
        elif isinstance(index_or_slice, slice):
            return self._program[index_or_slice]
        else:
            raise TypeError("Index must be an int or a slice.")

    def _single_char_repr(self) -> str:
        return "".join(command.value for command in self._program)

    def __repr__(self) -> str:
        return f"Program({self._single_char_repr()!r})"

    def __str__(self) -> str:
        return self._single_char_repr()

    def _move_right_nostep(self) -> None:
        """Move the program pointer to the right without incrementing the step counter."""
        if self._pointer < len(self) - 1:
            self._pointer += 1
        else:
            raise ProgramPointerError("Program pointer overflow.")

    def move_right(self, N: int = 1) -> None:
        """Move the program pointer to the right."""
        for _ in range(N):
            self._steps += 1
            self._move_right_nostep()

    def _move_left_nostep(self) -> None:
        """Move the program pointer to the left without incrementing the step counter."""
        if self._pointer > 0:
            self._pointer -= 1
        else:
            raise ProgramPointerError("Program pointer underflow.")

    def move_left(self, N: int = 1) -> None:
        """Move the program pointer to the left."""
        for _ in range(N):
            self._steps += 1
            self._move_left_nostep()

    def loop_start(self, current_bit: bool) -> None:
        """If the current bit is zero, jump past matching ). Else, continue (loop)."""
        if self.command != Command.LOOP_START:
            raise ValueError("Not at a loop start.")

        # NOTE: we want to increment even if we end up raising an error.
        self._steps += 1

        if not current_bit:
            try:
                # The current bit is zero, so we jump past the matching ).
                bracket_depth = 1
                while bracket_depth > 0:
                    self._move_right_nostep()
                    if self.command == Command.LOOP_START:
                        bracket_depth += 1
                    elif self.command == Command.LOOP_END:
                        bracket_depth -= 1

            except ProgramPointerError:
                # We've hit the end of the program without finding the matching ).
                raise InvalidProgramError("Unmatched loop start.")

        # We are now at the matching ), but we need to move past it, so move right once more.
        try:
            self._move_right_nostep()
        except ProgramPointerError:
            # We cannot move past the matching ) because we are at the end of the program.
            # Just let the program end by re-raising the ProgramPointerError.
            raise

    def loop_end(self, current_bit: bool) -> None:
        """If the current bit is zero, jump back to just after matching (. Else, continue."""
        if self.command != Command.LOOP_END:
            raise ValueError("Not at a loop end.")

        self._steps += 1

        if not current_bit:
            try:
                # The current bit is zero, so we jump back to just after the matching (.
                bracket_depth = 1
                while bracket_depth > 0:
                    self._move_left_nostep()
                    if self._program[self._pointer] == Command.LOOP_START:
                        bracket_depth -= 1
                    elif self._program[self._pointer] == Command.LOOP_END:
                        bracket_depth += 1

            except ProgramPointerError:
                raise InvalidProgramError("Unmatched loop end.")

        # We are now at the matching (, but we need to move just after it, so move right once.
        # NOTE: We don't need to worry about hitting the end of the program here.
        self._move_right_nostep()

    def reset(self) -> None:
        """Reset the program."""
        self._pointer = 0
        self._steps = 0

    @property
    def command(self) -> Command:
        if len(self) == 0:
            raise ProgramPointerError("Program is empty.")
        return self._program[self._pointer]

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Program):
            return self._program == other.program
        elif isinstance(other, str):
            return str(self) == other
        else:
            return False

    def __hash__(self) -> int:
        return hash(str(self))

    def copy(self) -> "Program":
        """Return a copy of the program."""
        return Program(self)


def preprocess_program(program: str) -> str:
    """Preprocess the RBF program."""

    if isinstance(program, str):
        # Split into lines
        lines = program.split("\n")
        # For each line, remove any characters after a # (comments)
        lines = [line.split("#", 1)[0] for line in lines]
        # Join the lines back together
        program = "".join(lines)
        # Remove spaces
        program = program.replace(" ", "")

    return program


def validate_program(program: Union[str, Sequence[Command]]) -> Sequence[Command]:
    """Validate the RBF program."""

    parsed_commands: Sequence[Command]
    if isinstance(program, str):
        # Remove any characters that are not RBF commands.
        # Convert the string to a list of Commands.
        parsed_commands = [Command(command) for command in program]
    else:
        # The program is already a list of Commands.
        parsed_commands = program

    bracket_depth = 0  # Keep track of the depth of the brackets.
    for command in parsed_commands:
        if command == Command.LOOP_START:
            bracket_depth += 1
        elif command == Command.LOOP_END:
            bracket_depth -= 1

        if bracket_depth < 0:
            raise InvalidProgramError("Unmatched loop end.")

    if bracket_depth != 0:
        raise InvalidProgramError("Unmatched loop start.")

    return parsed_commands



from .runner import run


def main() -> None:
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("source", help="The source code to run")
    parser.add_argument(
        "-t",
        "--tape",
        help="The initial tape to use. Can be a string of 1s and 0s, or an integer to use as the tape size",
        default="8",
    )
    parser.add_argument(
        "--max-steps",
        type=int,
        help="The maximum number of steps to run",
        default=10_000,
    )
    parser.add_argument(
        "--log-level",
        help="The log level to use",
        default="INFO",
    )

    args = parser.parse_args()

    import logging

    logger = logging.getLogger("rbf")

    logger.setLevel(args.log_level)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(levelname)05s %(name)s: %(message)s"))
    logger.addHandler(handler)

    # Check if tape is a string containing only 1s and 0s
    if all(x in "01" for x in args.tape):
        logger.debug("Using --tape as a string")
        tape = args.tape
    else:
        logger.debug("Using --tape as an integer")
        tape = int(args.tape)

    def callback(program, tape):
        logger.debug(
            f" {program.steps} {program.command.value} {program.pointer:02d} | {tape.pointer:02d} {tape}"
        )
        return False

    _program, tape = run(
        args.source,
        tape,
        max_steps=args.max_steps,
        callback=callback,
    )

    print(tape)



import pytest
from rbf_lang.program import Program, ProgramPointerError, InvalidProgramError

from rbf_lang.command import Command


def test_run():
    source = "*>" * 8
    program = Program(source)

    assert program == source


def test_hashable():
    source = "*>" * 8
    program = Program(source)
    hash_1 = hash(source)
    hash_2 = hash(program)

    assert hash_1 == hash_2


def test_equality():
    source = "*>" * 8
    program_1 = Program(source)
    program_2 = Program(source)

    assert program_1 == program_2


def test_move():
    source = "*>" * 8
    program = Program(source)

    assert len(program) == 16
    assert program.pointer == 0

    for i in range(15):
        assert program.pointer == i
        program.move_right()

    assert program.steps == 15

    # We now should be at the end of the program
    assert program.pointer == 15
    with pytest.raises(ProgramPointerError):
        program.move_right()

    assert program.pointer == 15
    # This still should have incremented the steps
    assert program.steps == 16

    for i in range(15, 0, -1):
        assert program.pointer == i
        program.move_left()

    assert program.pointer == 0
    assert program.steps == 31

    with pytest.raises(ProgramPointerError):
        program.move_left()

    assert program.steps == 32


def test_reset():
    source = "*>" * 8
    program = Program(source)

    program.move_right(4)

    assert program.pointer == 4
    assert program.steps == 4

    program.reset()

    assert program.pointer == 0
    assert program.steps == 0


def test_getitem():
    source = "*>" * 8
    program = Program(source)

    assert program[0] == Command("*")
    assert program[1] == Command(">")
    assert program[2] == Command("*")

    assert program[0:3] == [Command("*"), Command(">"), Command("*")]


def test_command_property():
    source = "*>" * 8
    program = Program(source)

    assert program.command == Command("*")
    program.move_right()
    assert program.command == Command(">")


def test_loop_start():
    source = "(*)*"
    program = Program(source)

    # Current bit is 0, so we should jump *past* the matching )
    program.loop_start(False)
    assert program.pointer == 3
    assert program.steps == 1

    program.reset()

    # Current bit is 1, so we should noop (but still increment steps)
    program.loop_start(True)
    assert program.pointer == 1
    assert program.steps == 1


def test_loop_start_overflow():
    source = "(*)"
    program = Program(source)

    # Current bit is 0, so we should jump *past* the matching )
    with pytest.raises(ProgramPointerError):
        program.loop_start(False)
    assert program.pointer == 2
    assert program.steps == 1

    program.reset()

    # Current bit is 1, so we should noop (but still increment steps)
    program.loop_start(True)
    assert program.pointer == 1
    assert program.steps == 1


def test_loop_start_only_on_bracket():
    source = "(*)"
    program = Program(source)
    program.move_right()

    # We're not on a loop start bracket
    with pytest.raises(ValueError):
        program.loop_start(False)


def test_loop_start_nested_brackets():
    source = "(*(*))*"
    program = Program(source)

    # Current bit is 0, so we should jump *past* the matching )
    program.loop_start(False)
    assert program.pointer == 6
    assert program.steps == 1


def test_loop_end():
    source = "(*)*"
    program = Program(source)
    program.move_right(2)

    # Current bit is 0, so we should jump back to just after the matching (
    program.loop_end(False)
    assert program.pointer == 1
    assert program.steps == 3

    program.reset()
    program.move_right(2)

    # Current bit is 1, so we should noop (but still increment steps)
    program.loop_end(True)
    assert program.pointer == 3
    assert program.steps == 3


def test_loop_end_overflow():
    source = "(*)>"
    program = Program(source)
    program.move_right(2)

    # Current bit is 0, so we should jump back to just after the matching (
    program.loop_end(False)
    assert program.pointer == 1
    assert program.steps == 3

    program.reset()
    program.move_right(2)

    # Current bit is 1, so we should noop (but still increment steps)
    program.loop_end(True)
    assert program.pointer == 3
    assert program.steps == 3


def test_loop_end_only_on_bracket():
    source = "(*)"
    program = Program(source)

    # We're not on a loop end bracket
    with pytest.raises(ValueError):
        program.loop_end(False)


def test_loop_end_nested_brackets():
    source = "(*(*))*"
    program = Program(source)
    program.move_right(5)

    program.loop_end(False)
    assert program.pointer == 1
    assert program.steps == 6


def test_invalid_program():
    source = "*)"
    with pytest.raises(InvalidProgramError):
        Program(source)


def test_copy():
    source = "*>" * 8
    program = Program(source)
    program_2 = program.copy()

    assert program == program_2
    assert program is not program_2



import pytest
from rbf_lang.tape import Tape


@pytest.fixture
def tape() -> Tape:
    return Tape(8)


def test_tape(tape: Tape) -> None:
    assert len(tape) == 8
    assert tape.pointer == 0
    assert tape.tape == [False] * 8


def test_getitem(tape: Tape) -> None:
    assert tape[0] is False
    assert tape[1] is False
    assert tape[0:2] == [False, False]


def test_copy(tape: Tape) -> None:
    tape_2 = tape.copy()
    assert tape == tape_2
    assert tape is not tape_2

    tape_3 = Tape(tape)
    assert tape == tape_3
    assert tape is not tape_3


def test_hashable(tape: Tape) -> None:
    tape_2 = tape.copy()
    assert hash(tape) == hash(tape_2)
    assert hash(tape) == hash("00000000")


def test_alternative_inits() -> None:
    tape = Tape([1, 0, 1, 0, 1, 0, 1, 0])
    assert tape == "10101010"
    assert tape == [1, 0, 1, 0, 1, 0, 1, 0]

    tape = Tape("10101010")
    assert tape == "10101010"
    assert tape == [1, 0, 1, 0, 1, 0, 1, 0]


def test_toggle(tape: Tape) -> None:
    assert tape[0] is False
    tape.toggle()
    assert tape[0] is True
    tape.toggle()
    assert tape[0] is False


def test_move(tape: Tape) -> None:
    assert len(tape) == 8
    assert tape.pointer == 0

    for i in range(7):
        assert tape.pointer == i
        tape.move_right()

    assert tape.pointer == 7

    tape.move_right()
    assert tape.pointer == 0

    tape.move_left()
    assert tape.pointer == 7
    for i in range(7, 0, -1):
        assert tape.pointer == i
        tape.move_left()


def test_reset(tape: Tape) -> None:
    tape.move_right(3)
    assert tape.pointer == 3
    tape.reset()
    assert tape.pointer == 0


def test_bit(tape: Tape) -> None:
    assert tape.bit is False
    tape.toggle()
    assert tape.bit is True
    tape.toggle()
    assert tape.bit is False



from typing import Callable, Optional
from .command import Command
from .program import Program, ProgramPointerError, _ProgramInitType
from .tape import Tape, _TapeInitType


def run(
    program: _ProgramInitType,
    tape: _TapeInitType,
    max_steps: int = 1000,
    callback: Optional[Callable[[Program, Tape], bool]] = None,
) -> tuple[Program, Tape]:
    """Run the RBF program. The program will run until it reaches the maximum number of steps or the callback returns True."""

    program = Program(program)
    tape = Tape(tape)
    try:
        while program.steps < max_steps:
            # Run the callback and break if it returns True
            if callback and callback(program, tape):
                break

            command = program.command
            bit = tape.bit

            if command == Command.TOGGLE:
                tape.toggle()
                program.move_right()
            elif command == Command.TAPE_RIGHT:
                tape.move_right()
                program.move_right()
            elif command == Command.TAPE_LEFT:
                tape.move_left()
                program.move_right()
            elif command == Command.LOOP_START:
                program.loop_start(bit)
            elif command == Command.LOOP_END:
                program.loop_end(bit)
            else:
                raise ValueError(f"Unknown command: {command}")

    except ProgramPointerError:
        # We've moved off the end of the program. This is fine.
        pass

    return program, tape



from rbf_lang.runner import run


def test_run_toggle() -> None:
    source = "*>" * 8
    tape_size = 8

    program, tape = run(
        source,
        tape_size,
    )

    assert program == source
    assert tape == "11111111"

    source = "*<" * 8
    program, tape = run(
        source,
        tape,
    )

    assert program == source
    assert tape == "00000000"


def test_run_callback() -> None:
    source = "*>" * 8
    tape_size = 8

    def callback(program, tape):
        return program.steps == 5

    program, tape = run(
        source,
        tape_size,
        callback=callback,
    )

    assert program.steps == 5
    assert tape == "11100000"


def test_run_max_steps() -> None:
    source = "*>" * 8
    tape_size = 8

    program, tape = run(
        source,
        tape_size,
        max_steps=5,
    )

    assert program.steps == 5
    assert tape == "11100000"


def test_run_loop_behavior() -> None:
    source = "()"
    program, tape = run(source, 8, max_steps=10)

    assert program.steps == 1
    assert tape == "00000000"

    source = "*(>)"
    program, tape = run(source, 8, max_steps=10)
    assert program.steps == 10



class RBFError(Exception):
    pass


class ProgramPointerError(RBFError):
    """Raised when the program counter is out of bounds."""


class InvalidProgramError(RBFError):
    """Raised when the program is invalid."""
